/*
 * multican 0.0.4, USB remote control tool for Canon cameras.
 *
 * Copyright (C) 2006 Jindrich Novy (jnovy@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <endian.h>
#include <time.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <usb.h>

//#define DEBUG 1
#define ABSOLUTE_TIME 1

#define USB_TIMEOUT	15000	/* miliseconds */
#define USB_MAX_DEVICES 6

#define USB_READ	0xc0
#define USB_WRITE	0x40
#define USB_SINGLE_BYTE	0x0c
#define USB_MORE_BYTES	0x04

#define USB_READ_BYTE(handle,value,byteptr) \
	usb_control_msg(handle, USB_READ, USB_SINGLE_BYTE, value, 0, byteptr, 1, USB_TIMEOUT)
#define USB_READ_BYTES(handle,value,bytes,size) \
	usb_control_msg(handle, USB_READ, size>1?USB_MORE_BYTES:USB_SINGLE_BYTE, value, 0, bytes, size, USB_TIMEOUT)
#define USB_WRITE_BYTE(handle,value,byteptr) \
	usb_control_msg(handle, USB_WRITE, USB_SINGLE_BYTE, value, 0, byteptr, 1, USB_TIMEOUT)
#define USB_WRITE_BYTES(handle,value,bytes,size) \
	usb_control_msg(handle, USB_WRITE, size>1?USB_MORE_BYTES:USB_SINGLE_BYTE, value, 0, bytes, size, USB_TIMEOUT)

#define CANON_CLASS(c) (camera_id[(c)->cam_id].Class)
#define IS_EOS(c) (CANON_CLASS(c) == 4 || CANON_CLASS(c) == 6)
#define IS_TYPE(c,s) (strstr(camera_id[(c)->cam_id].name, (s)) != NULL)

#ifdef DEBUG
#  define FUNCSTART fprintf(stderr, "%20s(): % 4d --------------------------------------------\n", __FUNCTION__, __LINE__);
#else
#  define FUNCSTART
#endif

static struct {
	unsigned int idVendor, idProduct, Class;
	char *name;
} camera_id[] = {
	{0x04a9, 0x3041, 0, "PowerShot S10"},
	{0x04a9, 0x3043, 0, "PowerShot S20"},
	{0x04a9, 0x3044, 4, "EOS D30"},
	{0x04a9, 0x3045, 0, "PowerShot S100"},
	{0x04a9, 0x3046, 0, "IXY Digital"},
	{0x04a9, 0x3047, 0, "Digital IXUS"},
	{0x04a9, 0x3048, 0, "PowerShot G1"},
	{0x04a9, 0x3049, 0, "PowerShot Pro90 IS"},
	{0x04a9, 0x304b, 1, "IXY Digital 300"},
	{0x04a9, 0x304c, 1, "PowerShot S300"},
	{0x04a9, 0x304d, 1, "Digital IXUS 300"},
	{0x04a9, 0x304e, 1, "PowerShot A20"},
	{0x04a9, 0x304f, 1, "PowerShot A10"},
	{0x04a9, 0x3050, 1, "PowerShot unknown 1"},
	{0x04a9, 0x3051, 0, "PowerShot S110"},
	{0x04a9, 0x3052, 0, "Digital IXUS V"},
	{0x04a9, 0x3055, 1, "PowerShot G2"},
	{0x04a9, 0x3056, 1, "PowerShot S40"},
	{0x04a9, 0x3057, 1, "PowerShot S30"},
	{0x04a9, 0x3058, 1, "PowerShot A40"},
	{0x04a9, 0x3059, 1, "PowerShot A30"},
	{0x04a9, 0x305b, 1, "ZR45MC Digital Camcorder"},
	{0x04a9, 0x3060, 4, "EOS D60"},
	{0x04a9, 0x3061, 1, "PowerShot A100"},
	{0x04a9, 0x3062, 1, "PowerShot A200"},
	{0x04a9, 0x3065, 1, "PowerShot S200 or Digital IXUS V2"},
	{0x04a9, 0x3066, 1, "PowerShot S330 or Digital IXUS 330"},
	{0x04a9, 0x3067, 1, "MV550i Digital Video Camera"},
	{0x04a9, 0x3069, 1, "PowerShot G3"},
	{0x04a9, 0x306a, 1, "Digital unknown 3"},
	{0x04a9, 0x306b, 1, "Optura 200 MC or IXY2i or MVX2i Digital Video Camera"},
	{0x04a9, 0x306c, 5, "PowerShot S45"},
	{0x04a9, 0x306d, 5, "PowerShot S45 (ptp)"},
	{0x04a9, 0x306e, 5, "PowerShot G3"},
	{0x04a9, 0x306f, 5, "PowerShot G3 (ptp)"},
	{0x04a9, 0x3070, 4, "PowerShot S230 or Digital IXUS V3"},
	{0x04a9, 0x3071, 4, "PowerShot S230 (ptp) or Digital IXUS V3 (ptp)"},
	{0x04a9, 0x3072, 5, "PowerShot SD100 or Digital IXUS II"},
	{0x04a9, 0x3073, 1, "PowerShot A70"},
	{0x04a9, 0x3074, 1, "PowerShot A60"},
	{0x04a9, 0x3075, 1, "Digital IXUS 400 Camera or PowerShot S400"},
	{0x04a9, 0x3076, 1, "PowerShot A300"},
	{0x04a9, 0x3077, 4, "PowerShot S50"},
	{0x04a9, 0x3078, 1, "ZR70MC Digital Camcorder"},
	{0x04a9, 0x3079, 1, "MV650i Digital Video Camera (ptp)"},
	{0x04a9, 0x307a, 1, "MV650i Digital Video Camera"},
	{0x04a9, 0x307b, 1, "MV630i Digital Video Camera (ptp)"},
	{0x04a9, 0x307c, 1, "MV630i Digital Video Camera"},
	{0x04a9, 0x307f, 1, "Optura 20 (ptp) or MVX150i (ptp)"},
	{0x04a9, 0x3080, 1, "Optura 20 or MVX150i"},
	{0x04a9, 0x3081, 1, "Optura 10 or MVX100i"},
	{0x04a9, 0x3082, 1, "Optura 10 or MVX100i"},
	{0x04a9, 0x3083, 4, "EOS 10D"},
	{0x04a9, 0x3084, 4, "EOS 300D or Digital Rebel or Kiss Digital"},
	{0x04a9, 0x3085, 5, "PowerShot G5"},
	{0x04a9, 0x3087, 1, "Elura 50 (ptp)"},
	{0x04a9, 0x3088, 1, "Elura 50"},
	{0x04a9, 0x308d, 1, "Optura Xi (ptp) or MVX 3i (ptp) or FV M1 (ptp)"},
	{0x04a9, 0x308e, 1, "Optura Xi or MVX 3i or FV M1"},
	{0x04a9, 0x3095, 1, "Optura 300 (ptp) or MVX 10i (ptp) or IXY DV M2 (ptp)"},
	{0x04a9, 0x3096, 1, "Optura 300 or MVX 10i or IXY DV M2"},
	{0x04a9, 0x3099, 4, "EOS 300D (ptp)"},
	{0x04a9, 0x309a, 1, "PowerShot A80"},
	{0x04a9, 0x309b, 1, "SD10 Digital ELPH or Digital IXUS i or IXY Digital L"},
	{0x04a9, 0x309c, 5, "PowerShot S1 IS"},
	{0x04a9, 0x30a9, 1, "Optura 40/MVX25i camcorder"},
	{0x04a9, 0x30b1, 5, "PowerShot S70"},
	{0x04a9, 0x30b2, 5, "PowerShot S60"},
	{0x04a9, 0x30b3, 5, "PowerShot G6"},
	{0x04a9, 0x30b4, 5, "Digital IXUS 500 or PowerShot S500 Digital ELPH"},
	{0x04a9, 0x30b5, 1, "PowerShot A75"},
	{0x04a9, 0x30b6, 1, "PowerShot SD110 Digital ELPH or Digital IXUS IIs"},
	{0x04a9, 0x30b7, 5, "PowerShot A400"},
	{0x04a9, 0x30b8, 5, "PowerShot A310"},
	{0x04a9, 0x30b9, 5, "Powershot A85"},
	{0x04a9, 0x30ba, 5, "PowerShot S410 Digital ELPH or Digital IXUS 430 or IXY Digital 430"},
	{0x04a9, 0x30bb, 5, "PowerShot A95"},
	{0x04a9, 0x30bf, 6, "PowerShot SD300 or Digital IXUS 40"}, /* IXUS likely in ptp */
	{0x04a9, 0x30c0, 6, "PowerShot SD200 or Digital IXUS 30 or IXY Digital 40"},
	{0x04a9, 0x30c1, 4, "PowerShot SD400 or Digital IXUS 50 or IXY Digital 55"},
	{0x04a9, 0x30c2, 1, "PowerShot A510"},
	{0x04a9, 0x30c4, 5, "PowerShot SD20 or Digital IXUS i5 or IXY Digital L2"},
	{0x04a9, 0x30e9, 6, "EOS 1D Mark II"},
	{0x04a9, 0x30ea, 6, "EOS 1D Mark II (ptp)"},
	{0x04a9, 0x30eb, 6, "EOS 20D"},
	{0x04a9, 0x30ec, 6, "EOS 20D (ptp)"},
	{0x04a9, 0x30ee, 6, "EOS 350D or Digital Rebel XT or EOS Kiss Digital N"},
	{0x04a9, 0x30ef, 6, "EOS 350D (ptp) or Digital Rebel XT (ptp) or EOS Kiss Digital N (ptp)"},
	{0x04a9, 0x30f0, 1, "PowerShot S2 IS"},
	{0x04a9, 0x30f2, 5, "PowerShot SD500 or Digital IXUS 700 or IXY Digital 600"},
	{0x04a9, 0x30fd, 5, "PowerShot A610"},
	{0x04a9, 0x3101, 6, "EOS 5D"},
	{0x04a9, 0x3110, 6, "EOS 400D"},
	{0x04a9, 0x3113, 6, "EOS 30D"},
	{0x04a9, 0x3114, 6, "EOS 30D (ptp)"},
	{0x04a9, 0x311a, 4, "PowerShot S3 IS"},
	{0, 0, 0, NULL}
};

static struct {
	unsigned int code;
	char *text;
} canon_status[] = {
	{0x02000022, "File not found"},
	{0x02000029, "File was protected"},
	{0x0200002a, "Compact Flash card full"},
	{0x02000081, "Failed to lock EOS keys"},
	{0x02000082, "Failed to unlock EOS keys"},
	{0x02000085, "Could not switch to capture mode"},
	{0x02000086, "Invalid command parameters"},
	{0x00000086, "Can't unlock EOS keys (new)"},
	{0x02000087, "No storage card in camera"},
	{0x82200040, "Unknown error (new protocol)"},
	{0x82220040, "Unknown error (new protocol)"},
	{0x00000000, NULL}
};

static struct {
	char *token;
	char code;
} tv_vals[] = {
	{ "bulb", 0x04 },
	{ "30\"", 0x10 },
	{ "25\"", 0x13 },
	{ "20\"", 0x15 },
	{ "15\"", 0x18 },
	{ "13\"", 0x1b },
	{ "10\"", 0x1d },
	{ "8\"", 0x20 },
	{ "6\"", 0x23 },
	{ "5\"", 0x25 },
	{ "4\"", 0x28 },
	{ "3\"2", 0x2b },
	{ "2\"5", 0x2d },
	{ "2\"", 0x30 },
	{ "1\"6", 0x33 },
	{ "1\"3", 0x35 },
	{ "1\"", 0x38 },
	{ "0\"8", 0x3b },
	{ "0\"6", 0x3d },
	{ "0\"5", 0x40 },
	{ "0\"4", 0x43 },
	{ "0\"3", 0x45 },
	{ "4", 0x48 },
	{ "5", 0x4b },
	{ "6", 0x4d },
	{ "8", 0x50 },
	{ "10", 0x53 },
	{ "13", 0x55 },
	{ "15", 0x58 },
	{ "20", 0x5b },
	{ "25", 0x5d },
	{ "30", 0x60 },
	{ "40", 0x63 },
	{ "50", 0x65 },
	{ "60", 0x68 },
	{ "80", 0x6b },
	{ "100", 0x6d },
	{ "125", 0x70 },
	{ "160", 0x73 },
	{ "200", 0x75 },
	{ "250", 0x78 },
	{ "320", 0x7b },
	{ "400", 0x7d },
	{ "500", 0x80 },
	{ "640", 0x83 },
	{ "800", 0x85 },
	{ "1000", 0x88 },
	{ "1250", 0x8b },
	{ "1600", 0x8d },
	{ "2000", 0x90 },
	{ "2500", 0x93 },
	{ "3200", 0x95 },
	{ "4000", 0x98 },
	{ NULL, 0 }
};

static struct {
	char *token;
	char code;
} av_vals[] = {
	{ "29", 0x55 },
	{ "25", 0x53 },
	{ "22", 0x50 },
	{ "20", 0x4d },
	{ "18", 0x4b },
	{ "16", 0x48 },
	{ "14", 0x45 },
	{ "13", 0x43 },
	{ "11", 0x40 },
	{ "10", 0x3d },
	{ "9.0", 0x3b },
	{ "8.0", 0x38 },
	{ "7.1", 0x35 },
	{ "6.3", 0x33 },
	{ "5.6", 0x30 },
	{ "5.0", 0x2d },
	{ "4.5", 0x2b },
	{ "4.0", 0x28 },
	{ "3.5", 0x25 },
	{ "3.2", 0x23 }, /* untested */
	{ "2.8", 0x20 }, /* untested */
	{ NULL, 0 }
};

static struct {
	char *token;
	char code;
} iso_vals[] = {
	{ "1600", 0x68 },
	{ "1250", 0x65 }, /* untested */
	{ "1000", 0x63 }, /* untested */
	{ "800", 0x60 },
	{ "640", 0x5d }, /* untested */
	{ "500", 0x5b }, /* untested */
	{ "400", 0x58 },
	{ "320", 0x55 }, /* untested */
	{ "250", 0x53 }, /* untested */
	{ "200", 0x50 },
	{ "160", 0x4d }, /* untested */
	{ "125", 0x4b }, /* untested */
	{ "100", 0x48 },
	{ NULL, 0}
};

int tolerate_error = 0, retries = 5000;

typedef struct {
	usb_dev_handle *h;
	struct usb_device *dev;
	struct usb_interface *interface;
	struct usb_interface_descriptor *desc;
	int bulk_in;
	int bulk_out;
	int bulk_int_in;
	int cam_id;
	char canon_state;
	char *camera_type;
	char *owner;
	char *firmware;
	int  firmware1;
	int  firmware2;
	int  firmware3;
	int  firmware4;
	int identified;
	char *power_status;
	char *power_type;
	unsigned int body_id;
	int got_body_id;
	int flash_state;
	int thumb_to_pc;
	int full_to_pc;
	int thumb_to_drive;
	int full_to_drive;
	unsigned int avail_shots;
	int ready;
	unsigned int max_transfer;
	char rel_params[0x30];
	int rel_params_filled;
} camera;

void hexdump( char *bytes, int size ) {
	int n, i;
	
	fputs("         0  1  2  3  4  5  6  7| 8  9  A  B  C  D  E  F 01234567|89ABCDEF\n", stderr);
	for ( n=0; n<size; n+=i ) {
		fprintf(stderr, "0x%04X: ", n);
		for ( i=0; n+i<size && i<0x10; i++ ) {
			unsigned char b = bytes[n+i];
			fprintf(stderr, "%02X%c", b, i!=7?' ':'|');
		}
		for ( ; i<0x10; i++ )
			fputs("   ", stderr);
		for ( i=0; n+i<size && i<0x10; i++ ) {
			unsigned b = bytes[n+i];
			if ( i==8 )
				putc('|', stderr);
			putc(b>=0x20 && b<0x7f ? b : '.', stderr);
		}
		putc('\n', stderr);
	}
	putc('\n', stderr);
}

#ifndef _GNU_SOURCE
char *strdup( char *s ) {
	size_t l = strlen(s)+1;
	char *r = malloc(l);
	
	memcpy(r, s, l);
	
	return r;
}
#endif

static unsigned get_le32( char *b ) {
	unsigned char A, B, C, D;
#if __BYTE_ORDER == __LITTLE_ENDIAN	
	A = b[0];
	B = b[1];
	C = b[2];
	D = b[3];
#elif __BYTE_ORDER == __BIG_ENDIAN
	A = b[3];
	B = b[2];
	C = b[1];
	D = b[0];
#else
#	error "Endianess of you machine is not supported."
#endif
	return A | (B<<8) | ((C)<<16) | ((D)<<24);
}

/* Works only for little endian machines at the time! */
static void put_le32( char *b, unsigned int v ) {
#if __BYTE_ORDER == __LITTLE_ENDIAN
	b[0] = v&0xff;
	b[1] = (v>>8)&0xff;
	b[2] = (v>>16)&0xff;
	b[3] = (v>>24)&0xff;
#elif __BYTE_ORDER == __BIG_ENDIAN
	b[3] = v&0xff;
	b[2] = (v>>8)&0xff;
	b[1] = (v>>16)&0xff;
	b[0] = (v>>24)&0xff;
#else
#	error "Endianess of you machine is not supported."
#endif
}

void usb_read_byte( usb_dev_handle *h, int value, char *byte ) {
	int ret;
#if DEBUG
	fprintf(stderr, "usb_read_byte: value=%d (0x%x)", value, value);
#endif
	while ( (ret=USB_READ_BYTE(h, value, byte)) == 0  ) {
#if DEBUG
		fprintf(stderr, "Read nothing, trying again.\n");
#endif
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Read failed, exiting...\n");
		exit(1);
	}
#if DEBUG
    fprintf(stderr,", byte=%c (0x%02x)\n", *(unsigned char*)byte, *(unsigned char*)byte);
#endif
}

void usb_read_bytes( usb_dev_handle *h, int value, char *bytes, int size ) {
	int ret;
#if DEBUG
	fprintf(stderr, "usb_read_bytes: value=%d (0x%x), size=%d (0x%x):\n", value, value, size, size);
#endif
	while ( (ret=USB_READ_BYTES(h, value, bytes, size)) < size  ) {
#if DEBUG
		fprintf(stderr, "Partial read, trying again.\n");
#endif
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Read failed, exiting...\n");
		exit(1);
	}
#if DEBUG
	hexdump(bytes, size);
#endif
}

void usb_write_byte( usb_dev_handle *h, int value, char *byte ) {
	int ret = 0, retr = 0;
#if DEBUG
	fprintf(stderr, "usb_write_byte: value=%d (0x%x), byte=%c (0x%02x)\n", value, value, *byte, *byte);
#endif

	while ( (ret=USB_WRITE_BYTE(h, value, byte)) == 0  ) {
#if DEBUG
		fprintf(stderr, "Wrote partially, trying again.\n");
#endif
		if ( tolerate_error && ++retr >= retries ) {
			fprintf(stderr,"Writing sequence tried %d times, giving up...\n", retr);
			return;
		}
	}

	if ( ret < 0 ) {
		fprintf(stderr, "Write failed, exiting...\n");
		exit(1);
	}
}

void usb_write_bytes( usb_dev_handle *h, int value, char *bytes, int size ) {
	int ret = 0, retr = 0;
#if DEBUG
	fprintf(stderr, "usb_write_bytes: value=%d (0x%x), size=%d (0x%x):\n", value, value, size, size);
	hexdump(bytes, size);
#endif
	while ( (ret=USB_WRITE_BYTES(h, value, bytes, size)) < size  ) {
#if DEBUG
		fprintf(stderr, "Wrote partially, trying again.\n");
#endif
		if ( tolerate_error && ++retr >= retries ) {
			fprintf(stderr,"Writing sequence tried %d times, giving up...\n", retr);
			return;
		}
	}
	
	if ( ret < 0 ) {
		fprintf(stderr, "Write failed, exiting...\n");
		exit(1);
	}
}

void usb_bulk_read_bytes( usb_dev_handle *h, int ep, char *bytes, int size ) {
	int ret = -1, pos = 0, retr = 0;

	if ( !size ) {
		fprintf(stderr, "warning: bulk_read request of 0 bytes!\n");
		return;
	}
	
	memset(bytes, 0, size);
	while ( pos != size && (ret = usb_bulk_read( h, ep, &bytes[pos], size-pos, USB_TIMEOUT )) >= 0 ) {
		pos += ret;
#if DEBUG
		fprintf(stderr, "usb_bulk_read: ep=%d (0x%x), size=%d (0x%x):\n", ep, ep, ret, ret);
#endif
		retr++;
		if ( tolerate_error && retr >= retries ) break;
	}
	
	if ( pos != size || ret < 0 ) {
		fprintf(stderr, "Error in usb_bulk_read() request.\n");
		if ( !tolerate_error ) {
			exit(1);
		} else {
			fprintf(stderr, "Tolerating the above error for now.\n");
		}
	}
#if DEBUG
	hexdump(bytes, size);
#endif
}

char packet[0x1400];
unsigned sequence = 0;

void usb_send_packet( camera *c, unsigned char cmd1, unsigned char cmd2, unsigned int cmd3, char *arguments, int argsize ) {
#ifdef DEBUG
	fprintf(stderr, "usb_send_packet: cmd1=%d (0x%x), cmd2=%d (0x%x), cmd3=%d (0x%x):\n", cmd1, cmd1, cmd2, cmd2, cmd3, cmd3);
#endif
	memset(packet, 0, 0x50);
	put_le32(packet, 0x10+argsize);
	put_le32(packet+0x48, 0x10+argsize);
	packet[0x40] = 0x2;
	packet[0x44] = cmd1;
	packet[0x47] = cmd2;
	put_le32(packet+4, cmd3);

	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		packet[0x46] = cmd3==0x202 ? 0x20 : 0x10;
	}

	put_le32(packet+0x4c, sequence++);

	if ( argsize<0 || argsize+0x50 > sizeof(packet)) {
		fprintf(stderr, "Invalid argsize = %d (0x%x).\n\n", argsize, argsize);
		exit(1);
	}

	if ( argsize > 0 ) {
		memcpy(packet+0x50, arguments, argsize);
	}

	usb_write_bytes(c->h, 0x10, packet, 0x50+argsize);
}

void usb_receive_packet( camera *c, char *bytes, int size ) {
#ifdef DEBUG
	fprintf(stderr, "usb_receive_packet: ep=%d (0x%x), size=%d (0x%x):\n", c->bulk_in, c->bulk_in, size, size);
#endif
	usb_bulk_read_bytes(c->h, c->bulk_in, bytes, size);

	{
		int i;
		unsigned int code = get_le32(&bytes[0x50]);
#ifdef DEBUG
		unsigned int cmd3 = ((bytes[5]-1)<<8)+bytes[4];

		if (cmd3==0x202) { /* Format of response block is different in case cmd3==0x202*/
			fprintf(stderr,"usb_receive_packet sanity check: unformatted binary data of size=%d follows\n", get_le32(&bytes[6]));
		} else {
			fprintf(stderr,"usb_receive_packet sanity check: cmd1=0x%02x, cmd2=0x%02x, cmd3=0x%03x, 2==%d, size=0x%04x\n",
			bytes[0x44], 
			bytes[0x47]-0x10,
			cmd3,
			bytes[0x40],
			get_le32(&bytes[0x48])+0x40);
		}
#endif
		if ( !code ) return;

		for ( i=0; canon_status[i].code; i++ )
			if ( canon_status[i].code == code ) {
				fprintf(stderr, "usb_receive_packet status code: 0x%08x -> %s.\n", code, canon_status[i].text);
				return;
			}
		fprintf(stderr, "usb_receive_packet exit code: 0x%08x -> UNKNOWN !!!\n", code);
	}
}

/*
 * Purpose:
 * - scan USB ports for attached Canon devices
 *
 * Returns:
 * - number of successfully recognized Canon devices attached to USB ports
 * - filled camera[6] field
 */
int canon_scan( camera **c ) {
	struct usb_bus *busses;
	struct usb_bus *bus;
	struct usb_interface_descriptor *desc;
	int count = 0, n, ready = 0;
	camera *cam;

	FUNCSTART

	usb_init();
	usb_find_busses();
	usb_find_devices();

	busses = usb_get_busses();

	*c = calloc(USB_MAX_DEVICES, sizeof(camera));
	cam = *c;

	for (bus = busses; bus; bus = bus->next) {
                struct usb_device *dev;
		int i;

		for (dev = bus->devices; dev; dev = dev->next) {
			if (dev->descriptor.idVendor || dev->descriptor.idProduct) {
#ifdef DEBUG
				printf("Found: %04x:%04x\n", dev->descriptor.idVendor, dev->descriptor.idProduct);
#endif
				for ( i=0; camera_id[i].idVendor; i++ )
					if ( dev->descriptor.idVendor == camera_id[i].idVendor && dev->descriptor.idProduct == camera_id[i].idProduct ) {
#ifdef DEBUG
						printf(" -> %s.\n", camera_id[i].name);
#endif
						cam[count].dev = dev;
						cam[count].cam_id = i;
						
						cam[count].h = usb_open(dev);
						if ( !cam[count].h ) {
							fprintf(stderr, "  usb_open(): Unable to open USB device.\n");
							count++;
							break;
						}

						if ( usb_claim_interface(cam[count].h, 0) < 0 ) {
							fprintf(stderr, "  usb_claim_interface(): Cannot claim USB interface.\n");
							count++;
							break;
						}
						
						if ( usb_reset(cam[count].h) ) {
							fprintf(stderr, "  usb_reset(): Unable to reset interface.\n");
							count++;
							break;
						}
						
						cam[count].interface = cam[count].dev->config->interface;
						if (cam[count].interface->num_altsetting != 1) {
							fprintf(stderr,"  bad altsetting (%d).\n", cam[count].interface->num_altsetting);
							count++;
							break;
						}
						
						cam[count].desc = &cam[count].interface->altsetting[0];
						desc = cam[count].desc;

						for (n = 0; n<desc->bNumEndpoints; n++) {
							struct usb_endpoint_descriptor *endpoint = &desc->endpoint[n];

							switch(endpoint->bmAttributes & USB_ENDPOINT_TYPE_MASK) {
								case USB_ENDPOINT_TYPE_BULK:
									if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN) {
										cam[count].bulk_in = endpoint->bEndpointAddress;
									} else {
										cam[count].bulk_out = endpoint->bEndpointAddress;
									}
								break;
								case USB_ENDPOINT_TYPE_INTERRUPT:
									if ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN) {
										cam[count].bulk_int_in = endpoint->bEndpointAddress;
									}
								break;
							}
        					}
#ifdef DEBUG
						fprintf(stderr, "  bulk_in=%d (0x%x), bulk_out=%d (0x%x), bulk_int_in=%d (0x%x)\n\n",
						cam[count].bulk_in, cam[count].bulk_in, cam[count].bulk_out, cam[count].bulk_out, cam[count].bulk_int_in, cam[count].bulk_int_in);
#endif
						if (cam[count].bulk_in == 0 || cam[count].bulk_out == 0 || cam[count].bulk_int_in == 0) {
							fprintf(stderr, "  bad endpoints.\n");
							count++;
							break;
						}

						cam[count].ready = 1;
						
						count++;
						ready++;
						if ( count>=7 ) {
							fprintf(stderr, "  maximal number of supported devices reached.\n");
							return ready;
						}
						break;
					}
                        }
                }
        }
	
	if (!ready) {
		free(cam);
		*c = NULL;
		return 0;
	}

	return ready;
}

/*
 * Purpose:
 * - initialize detected camera for communication
 *
 * Returns:
 * - 1 == ok
 * - 0 == error
 * - sets camera ready flag to 0 when initialization is unsuccessful
 * - sets max_transfer
 */
int canon_init_camera( camera *c ) {
	char buf[0x58];
	char canon_state;

	FUNCSTART
	
	if ( !c->ready ) return 0;
#ifdef DEBUG
	fprintf(stderr, "Initializing %s...\n", camera_id[c->cam_id].name);
#endif

	usb_read_byte(c->h, 0x55, &canon_state);
	c->canon_state = canon_state;

	switch( canon_state ) {
		case 'A':
#ifdef DEBUG
			printf("Camera was already active.\n\n");
#endif
			usb_read_bytes(c->h, 1, buf, sizeof(buf));
			usb_read_bytes(c->h, 4, buf, 0x50); /* XXX: requesttype 0x4 instead of 0x40? */

/*			if ( CANON_CLASS(c) >= 6 ) {
				char tmp[1024];
				usb_send_packet(c, 0x24, 0x12, 0x201, NULL, 0);
				usb_receive_packet(c, tmp, sizeof(tmp));
			}
*/
			break;
		case 'C':
#ifdef DEBUG
			printf("Camera was woken up.\n\n");
#endif
			usb_read_bytes(c->h, 1, buf, sizeof(buf));
#ifdef DEBUG
			fprintf(stderr,"Max download transfer length: %u (0x%x).\n\n", get_le32(buf+0x4c), get_le32(buf+0x4c));
#endif
			c->max_transfer = get_le32(buf+0x4c);
			buf[0] = 0x10;
			memmove(&buf[0x40], &buf[sizeof(buf)-0x10], 0x10);
			usb_write_bytes(c->h, 0x11, buf, 0x50);
			if ( CANON_CLASS(c) < 6 ) {
				usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x40);
#ifdef DEBUG
				fprintf(stderr, "Length should be 4, length read from packet is: %u (0x%x).\n\n", get_le32(buf), get_le32(buf));
#endif
				usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x4);
				usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			} else {
				if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D")) {
#ifdef DEBUG
					fprintf(stderr, "Class >= 6 camera reading all 0x44 bytes.\n\n");
#endif
					usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x44);
				} else {
					usb_bulk_read_bytes(c->h, c->bulk_in, buf, 0x40);
					usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
				}
			}
			break;
		case 'I':
		case 'E':
#ifdef DEBUG
			fprintf(stderr, "Camera reports an error.\n");
#endif
			c->ready = 0;
			return 0;
			break;
		default:
			fprintf(stderr, "Camera returned unknown state '%c' (0x%02x).\n", canon_state, canon_state);
			c->ready = 0;
			return 0;
	}
	
	return 1;
}

/*
 * Purpose:
 * - initialize all detected cameras.
 *
 * Returns:
 * - cameras successfuly initialized
 */
int canon_init( camera *c ) {
	int i, ready = 0;

	FUNCSTART

	for ( i=0; i<USB_MAX_DEVICES; i++ )
		if ( c[i].ready ) {
			canon_init_camera( &c[i] );
			if ( c[i].ready ) ready++;
		}

	return ready;
}

void canon_identify_camera( camera *c ) {
	char buf[0x9c];

	FUNCSTART

	if ( !c->identified )
		c->identified = 1;
	else
		return;

	usb_send_packet(c, 0x1, 0x12, 0x201, NULL, 0);
	usb_receive_packet(c, buf, sizeof(buf));

	printf("Camera type:\t\t'%s'.\nOwner:\t\t\t'%s'.\nFirmware version:\t%c.%c.%c.%c.\n",
		buf+0x5c, buf+0x7c, buf[0x5b]+'0', buf[0x5a]+'0', buf[0x59]+'0', buf[0x58]+'0');

	{
		char firmware[] = {"0.0.0.0"};
	
		c->camera_type = strdup(buf+0x5c);
		c->owner = strdup(buf+0x7c);

		firmware[0] = buf[0x5b]+'0';
		firmware[2] = buf[0x5a]+'0';
		firmware[4] = buf[0x59]+'0';
		firmware[6] = buf[0x58]+'0';
		c->firmware = strdup(firmware);
		
		c->firmware1 = buf[0x5b];
		c->firmware2 = buf[0x5a];
		c->firmware3 = buf[0x59];
		c->firmware4 = buf[0x59];
	}
}

void canon_power_status( camera *c ) {
	char buf[0x58];

	FUNCSTART
	
	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"1D") ) {
		usb_send_packet(c, 0xa, 0x12, 0x201, NULL, 0);
	} else {
		usb_send_packet(c, 0x13, 0x12, 0x201, NULL, 0);
	}
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->power_status = strdup(buf[0x54]==6?"good":buf[0x54]==4?"bad":"unknown");
	c->power_type   = strdup(buf[0x57]&0x20?"battery":"AC adapter");
	
	printf("Power status:\t\t%s.\nPower type:\t\t%s.\n", 
		c->power_status, c->power_type);

	if ( CANON_CLASS(c) == 6 )
		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
}

void canon_eos_id( camera *c ) {
	char buf[0x58];

	FUNCSTART
	
	if ( !c->got_body_id )
		c->got_body_id = 1;
	else
		return;
	
	if ( !IS_EOS(c) ) {
#ifdef DEBUG
		fprintf(stderr, "Unable to get body ID on non-EOS cameras.\n");
#endif
		exit(1);
	}
	
	if ( CANON_CLASS(c) >= 6 /*&& !IS_TYPE(c,"1D")*/ ) {
		usb_send_packet(c, 0x23, 0x12, 0x201, NULL, 0);
	} else {
		usb_send_packet(c, 0x1d, 0x12, 0x201, NULL, 0);
	}
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->body_id = get_le32(&buf[0x54]);
	
	printf("Serial number:\t\t%010u.\n", c->body_id);
}

void canon_init_control( camera *c ) {
	char buf[0x5c];
	char args[0x9];

	FUNCSTART

	if ( IS_TYPE(c,"300D") || IS_TYPE(c,"1D") ) {
		memset(args, 0, 8);
		usb_send_packet( c, 0x13, 0x12, 0x201, args, 8 );
		usb_receive_packet(c, buf, sizeof(buf));
	} else {
		memset(args, 0, 9);
		usb_send_packet( c, 0x25, 0x12, 0x201, args, 9 );
		usb_receive_packet(c, buf, sizeof(buf));
	}
}

/* Doesn't work with EOS 300D */
#define FLASH_OFF  0x00
#define FLASH_ON   0x01
#define FLASH_AUTO 0x02
void canon_flash( camera *c, unsigned char flags ) {
	char buf[0x5c];
	char args[0x38];

	FUNCSTART
	
	printf("Flash is set:\t\t");
	switch( flags ) {
		case FLASH_OFF:
			printf("OFF\n");
		break;
		case FLASH_ON:
			printf("ON\n");
		break;
		case FLASH_AUTO:
			printf("AUTO\n");
		break;
		default:
			printf("to unknown state, THIS IS A BUG!\n");
			return;
		break;
	}
	memset(args, 0, sizeof(args));
	args[0]    = 7;
	args[0x12] = flags;
	
	c->flash_state = flags;

	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
}

#define THUMB_TO_PC 0x0001
#define FULL_TO_PC 0x0002
#define THUMB_TO_DRIVE 0x0004
#define FULL_TO_DRIVE 0x0008
void canon_transfer( camera *c, unsigned char flags ) {
	char buf[0x5c];
	char args[0xc];
	
	FUNCSTART
	
	if ( flags&~15 ) {
		fprintf(stderr, "canon_transfer(): Flags not in range!\n");
	}
	
	c->thumb_to_pc = flags&THUMB_TO_PC;
	c->full_to_pc = flags&FULL_TO_PC;
	c->thumb_to_drive = flags&THUMB_TO_DRIVE;
	c->full_to_drive = flags&FULL_TO_DRIVE;
	
	printf("Transfer mode is:%s%s%s%s.\n",
		c->thumb_to_pc?"\tTHUMB_TO_PC":"",
		c->full_to_pc?"\tFULL_TO_PC":"",
		c->thumb_to_drive?"\tTHUMB_TO_DRIVE":"",
		c->full_to_drive?"\tFULL_TO_DRIVE":""
	      );

	memset(args, 0, sizeof(args));

	args[0] = 9;
	args[4] = 4;
	args[8] = flags;

	if ( IS_TYPE(c,"300D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
}

/* Doesn't work on EOS 300D */
void canon_get_zoom( camera *c ) {
	char buf[0x60];
	char args[0x8];

	FUNCSTART

	args[0] = 0xb;
	memset(args+1, 0, 7);
	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
	
	printf("Zoom:\t\t%d\n", buf[0x5e]);
}

void canon_avail_shots( camera *c ) {
	char buf[0x60];
	char args[0x8];
	
	FUNCSTART

	args[0] = 0xd;
	memset(args+1, 0, 7);
	usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	usb_receive_packet(c, buf, sizeof(buf));
	
	c->avail_shots = get_le32(&buf[0x5c]);
	printf("Available shots:\t%u\n", c->avail_shots);
}

void canon_exit_control( camera *c ) {
	char buf[0x5c];
	char args[0x8];

	FUNCSTART

	args[0] = 1;
	memset(args+1, 0, 7);

	if ( IS_TYPE(c,"300D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
}

void canon_close( camera **cptr ) {
	int i;

	FUNCSTART
	
	if ( !*cptr ) return;
	
	for ( i=0; i<USB_MAX_DEVICES; i++ ) {
		camera *c = &(*cptr)[i];
	
		free(c->owner);
		free(c->camera_type);
		free(c->firmware);
		free(c->power_status);
		free(c->power_type);
		
		if ( c->interface ) usb_release_interface(c->h, 0);
		if ( c->h ) usb_close(c->h);
	}
	
	free( *cptr );
	*cptr = NULL;
}

void canon_release_shutter( camera *c ) {
	char buf[0x5c];
	char args[0x9];

	FUNCSTART

	args[0] = 4;
	memset(args+1, 0, 8);

	if ( IS_TYPE(c,"300D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, 8 );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, 9 );

	usb_receive_packet(c, buf, sizeof(buf));
	
/*	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		char buf2[0x54];
		char args2[4];

		args2[0] = 0xf;
		memset(args2+1, 0, 3);
		usb_send_packet( c, 0x22, 0x12, 0x201, args2, 4 );
		usb_receive_packet(c, buf2, sizeof(buf2));
	}
*/
	usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);

	if ( c->thumb_to_pc ) {
		fprintf(stderr, "  THUMB_TO_PC not implemented!\n");
	}

	if ( c->full_to_pc ) {
		unsigned int index, size;

		if ( buf[0xc] != 0xc ) {
			fprintf(stderr,"Warning: FULL_TO_PC interrupt read error.\n");
		}

		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x17);

		index = get_le32(buf+0xc);
		size  = get_le32(buf+0x11);
		printf("  size = %u B\n", size);
		
		if ( CANON_CLASS(c) != 6 ) /* 20D */
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
		
		{
			char recv[0x40];
			char args[0x10];
			char *img;
			unsigned int read = 0, read_chunk, actually_read;
			
			put_le32(args, 0);
//			put_le32(&args[4], c->max_transfer);
			put_le32(&args[4], 0x1400);
			put_le32(&args[8], 2); /* 1 - thumbnail, 2 - full image  */
			put_le32(&args[0xc], index);

			usb_send_packet( c, 0x17, 0x12, 0x202, args, sizeof(args) );
			usb_receive_packet(c, recv, sizeof(recv));

			size = get_le32(&recv[6]);
			printf("  size = %u B\n", size);
			
			img = calloc(size, 1);
//			usb_bulk_read_bytes(c->h, c->bulk_in, img, 0x1400);
//			usb_receive_packet(c, img, size);

			do {
				if ( size - read > 0x1400 )
					read_chunk = 0x1400;
				else if ( size - read > 0x40 )
					read_chunk = (size - read) / 0x40 * 0x40;
				else
					read_chunk = size - read;

				actually_read = usb_bulk_read(c->h, c->bulk_in, img + read, read_chunk, USB_TIMEOUT);
//				usb_bulk_read_bytes(c->h, c->bulk_in, img + read, read_chunk);

				if ( actually_read != read_chunk ) {
					fprintf(stderr, "Read error!\n");
					exit(1);
				}

				read += read_chunk;
			} while ( read < size );

			{
				FILE *f = fopen("img.jpg","wb");
				fwrite(img, size, 1, f);
				fclose(f);
			}

			free(img);
		}
	}

	if ( c->full_to_drive )	 {
		if ( buf[0xc] != 0x1c && IS_TYPE(c,"300D")) {
			fprintf(stderr,"Warning: FULL_TO_DRIVE first interrupt read error.\n");
		}

		if ( IS_TYPE(c,"300D") ) {
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			if ( buf[0xc] != 0x1d ) {
				fprintf(stderr,"Warning: FULL_TO_DRIVE second interrupt read error.\n");
			}
		}
		
		if ( IS_TYPE(c,"300D") ) {
			usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);
			if ( buf[0x4] != 0xe ) {
				fprintf(stderr,"Warning: FULL_TO_DRIVE error saving image to CF.\n");
			}
		}
	}

	if ( IS_EOS(c) && !IS_TYPE(c,"300D") && !IS_TYPE(c,"350D") && !IS_TYPE(c,"5D") && !IS_TYPE(c,"20D"))
		usb_bulk_read_bytes(c->h, c->bulk_int_in, buf, 0x10);

/*	if ( CANON_CLASS(c) >= 6 && !IS_TYPE(c,"1D") ) {
		char buf2[0x54];
		char args2[4];

		args2[0] = 0xf;
		memset(args2+1, 0, 3);
		usb_send_packet( c, 0x21, 0x12, 0x201, args2, 4 );
		usb_receive_packet(c, buf2, sizeof(buf2));
	}
*/
	c->rel_params_filled = 0;
}

void canon_set_owner( camera *c, char *name ) {
	usb_send_packet( c, 0x1, 0x12, 0x201, name, strlen(name)+1 );
}

//char *iso;

/*
 * Has to be called before changing any of Tv, Av, ISO.
 */
void canon_get_rel_params( camera *c ) {
	char buf[0x8c];
	char args[0x8];

	FUNCSTART
	
	memset(args, 0, sizeof(args));
	args[0]    = 0x0a;

	if ( IS_TYPE(c,"300D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args, sizeof(args) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args, sizeof(args) );

	usb_receive_packet(c, buf, sizeof(buf));
	
	memcpy(c->rel_params, &buf[0x5c], 0x30);
	c->rel_params_filled = 1;
	
/*	{
		FILE *f = fopen("iso.h","at");
		fprintf(f,"\t{ \"%s\", 0x%02x },\n", iso, buf[0x78]);
		fclose(f);
	}*/
}

void canon_set_rel_params( camera *c ) {
	char buf2[0x5c];
	char args2[0x38];

	if (!c->rel_params_filled) {
		fprintf(stderr, "warning: Release parameters not previously read.\n");
		return;
	}

	memset(args2, 0, sizeof(args2));
	args2[0] = 7;
	args2[4] = 0x30;
	memcpy(&args2[8], c->rel_params, 0x30);

	if ( IS_TYPE(c,"300D") )
		usb_send_packet( c, 0x13, 0x12, 0x201, args2, sizeof(args2) );
	else
		usb_send_packet( c, 0x25, 0x12, 0x201, args2, sizeof(args2) );

	usb_receive_packet(c, buf2, sizeof(buf2));
}

void canon_tv( camera *c, char *token ) {
	int i;

	FUNCSTART
	
	if (!c->rel_params_filled)
		canon_get_rel_params(c);

	for ( i=0; tv_vals[i].token; i++ ) {
		if ( !strcmp(tv_vals[i].token, token) ) {
#ifdef DEBUG
			fprintf(stderr, "canon_tv: setting tv=%s.\n", tv_vals[i].token);
#endif
			c->rel_params[0x1e] = tv_vals[i].code; /* 0x7a */
			break;
		}
	}

	if ( !tv_vals[i].token )
		fprintf(stderr,"warning: tv=%s -> no such token -> ignoring !!!\n", token);
}

void canon_av( camera *c, char *token ) {
	int i;

	FUNCSTART
	
	if (!c->rel_params_filled)
		canon_get_rel_params(c);

	for ( i=0; av_vals[i].token; i++ ) {
		if ( !strcmp(av_vals[i].token, token) ) {
#ifdef DEBUG
			fprintf(stderr, "canon_av: setting av=%s.\n", av_vals[i].token);
#endif
			c->rel_params[0x1c] = av_vals[i].code; /* 0x78 */
			break;
		}
	}

	if ( !av_vals[i].token )
		fprintf(stderr,"warning: av=%s -> no such token -> ignoring !!!\n", token);
}

void canon_iso( camera *c, char *token ) {
	int i;

	FUNCSTART
	
	if (!c->rel_params_filled)
		canon_get_rel_params(c);

	for ( i=0; iso_vals[i].token; i++ ) {
		if ( !strcmp(iso_vals[i].token, token) ) {
#ifdef DEBUG
			fprintf(stderr, "canon_iso: setting iso=%s.\n", iso_vals[i].token);
#endif
			c->rel_params[0x1a] = iso_vals[i].code;  /* 0x76 */
			break;
		}
	}

	if ( !iso_vals[i].token )
		fprintf(stderr,"warning: iso=%s -> no such token -> ignoring !!!\n", token);
}

int scr_line;
char *scr, *scr_name;
off_t scr_size, scr_pos;

char *next_line() {
	while (scr_pos<scr_size && (scr[scr_pos] == '#' || scr[scr_pos] == '\n') ) {
		for (; scr_pos<scr_size && scr[scr_pos] != '\n'; scr_pos++);
		scr_pos++;
		scr_line++;
	}

	if (scr_pos<scr_size) {
		return &scr[scr_pos];
	}

	return NULL;
}

char *skip_line() {
	while (scr_pos<scr_size && scr[scr_pos] != '\n') scr_pos++;

	return next_line();
}

off_t skip_text() {
	off_t opos = scr_pos;

	while (scr_pos<scr_size && scr[scr_pos] != ' ' && scr[scr_pos] != '\t' && scr[scr_pos] != '\n') scr_pos++;

	return scr_pos-opos;
}

off_t skip_spaces() {
	off_t opos = scr_pos;

	while (scr_pos<scr_size &&
	       (scr[scr_pos] == ' ' || scr[scr_pos] == '\t') && scr[scr_pos] != '\n' && scr[scr_pos] != '\r') scr_pos++;

	return scr_pos-opos;
}

off_t scr_textsize() {
	off_t sz = skip_text();
	scr_pos -= sz;
	return sz;
}

void scr_err( char *message ) {
	fprintf(stdout, "%s:\tError at %d: %s.\n", scr_name, scr_line, message);
	exit(1);
}

void scr_warn( char *message ) {
	fprintf(stdout,"%s:\tWarning at %d: %s.\n", scr_name, scr_line, message);
}

void scr_tv( camera *c ) {
	skip_spaces();

	if ( scr[scr_pos] != '=' ) {
		scr_warn("'=' expected, skipping..");
		return;
	}

	scr_pos++;

	skip_spaces();

	{
		off_t size = scr_textsize();
		char buf[0x10];
		
		if ( size>sizeof(buf)-1 ) {
			scr_warn("Tv value definition too long, skipping..");
			return;
		}
		
		memcpy(buf, &scr[scr_pos], size);
		buf[size] = '\0';
		canon_tv(c, buf);
		canon_set_rel_params(c);
	}
}

void scr_av( camera *c ) {
	skip_spaces();

	if ( scr[scr_pos] != '=' ) {
		scr_warn("'=' expected, skipping..");
		return;
	}

	scr_pos++;

	skip_spaces();

	{
		off_t size = scr_textsize();
		char buf[0x10];
		
		if ( size>sizeof(buf)-1 ) {
			scr_warn("Av value definition too long, skipping..");
			return;
		}
		
		memcpy(buf, &scr[scr_pos], size);
		buf[size] = '\0';
		canon_av(c, buf);
		canon_set_rel_params(c);
	}
}

void scr_iso( camera *c ) {
	skip_spaces();

	if ( scr[scr_pos] != '=' ) {
		scr_warn("'=' expected, skipping...");
		return;
	}

	scr_pos++;

	skip_spaces();

	{
		off_t size = scr_textsize();
		char buf[0x10];
		
		if ( size>sizeof(buf)-1 ) {
			scr_warn("ISO value definition too long, skipping..");
			return;
		}
		
		memcpy(buf, &scr[scr_pos], size);
		buf[size] = '\0';
		canon_iso(c, buf);
		canon_set_rel_params(c);
	}
}

void scr_shot( camera *c ) {
	canon_release_shutter(c);
}

static struct {
	char *token;
	void (*cback)(camera *);
} scr_cback[] = {
	{ "tv", scr_tv },
	{ "av", scr_av },
	{ "iso", scr_iso },
	{ "shot", scr_shot },
	{ "nop", NULL },
	{ NULL, NULL }
};

#ifdef ABSOLUTE_TIME
struct timeval start_tv;
#endif

void process_script( camera *c ) {
	int i;

	FUNCSTART
	
	scr_pos = 0;
	scr_line = 1;

	if ( !next_line() )
		scr_err("Unexpected EOF");

	canon_init_control(c);

	{
		int h, m, s, u;
		struct timeval tv1, tv2;
#ifndef ABSOLUTE_TIME
		struct timeval tv_tmp;
#endif

		while ( 1 ) {

			if ( sscanf(&scr[scr_pos], "%02d:%02d:%02d.%03d", &h, &m, &s, &u) != 4 ) {
				scr_warn("Invalid time format, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			if ( h<0 || m<0 || s<0 || u<0 ) {
				scr_warn("Time values cannot be negative, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			if ( h>23 || m>59 || s>59 || u>999 ) {
				scr_warn("At least on time value ot of range, skipping..");
				if ( !skip_line() ) break;
				continue;
			}

			skip_text();
			skip_spaces();

			for ( i=0; scr_cback[i].token; i++ ) {
				if (!memcmp(scr_cback[i].token, &scr[scr_pos], strlen(scr_cback[i].token))) {
					scr_pos += strlen(scr_cback[i].token);
#ifdef ABSOLUTE_TIME
					{
						int n = scr_pos;

						gettimeofday(&tv1, NULL);
						tv2.tv_sec   = start_tv.tv_sec + h*3600 + m*60 + s + ((start_tv.tv_usec + u*1000) / 1000000);
						tv2.tv_usec  = (start_tv.tv_usec + u*1000) % 1000000;

/*						printf("---\nstv: %ld:%06ld\n", start_tv.tv_sec, start_tv.tv_usec);
						printf("tv1: %ld:%06ld -> %02d:%02d:%02d.%03d\n", tv1.tv_sec, tv1.tv_usec, h, m, s, u);
						printf("tv2: %ld:%06ld\n", tv2.tv_sec, tv2.tv_usec);
*/
//							while( nanosleep(&nano_tv, &nano_tv) == -1);
//							usleep((nano_tv.tv_sec-tv1.tv_sec)*1000000);
						if ( tv2.tv_sec >= tv1.tv_sec ) {
							struct timespec ntv;
							if ( tv2.tv_sec == tv1.tv_sec ) {
								ntv.tv_sec  = 0;
								ntv.tv_nsec = (tv2.tv_usec - tv1.tv_usec) * 1000;

								if ( ntv.tv_nsec > 0 ) {
//									printf("%ld:%09ld\t", ntv.tv_sec, ntv.tv_nsec);
									while( nanosleep(&ntv, &ntv) == -1);
								}
							} else {
								long long dt = (1000000-tv1.tv_usec)+tv2.tv_usec+1000000*((tv2.tv_sec-tv1.tv_sec)-1);
								ntv.tv_sec  = dt / 1000000;
								ntv.tv_nsec = (dt % 1000000) * 1000;

//								printf("%ld:%09ld\t", ntv.tv_sec, ntv.tv_nsec);
								while( nanosleep(&ntv, &ntv) == -1);
							}
						}

						fputs(scr_name,stdout);
						putchar('\t');
						printf("%02d:%02d:%02d.%03d\t", h, m, s, u);
						fputs(scr_cback[i].token, stdout);
						for ( ; n<scr_size && scr[n] != '\n'; n++ )
							putchar(scr[n]);
						putchar('\n');

						if ( scr_cback[i].cback ) scr_cback[i].cback(c);
					}
#else /* ABSOLUTE_TIME */
					{
						gettimeofday(&tv1, NULL);
						tv2.tv_sec  = tv1.tv_sec + h*3600 + m*60 + s + ((tv1.tv_usec + u*1000) > 1000000 ? 1 : 0);
						tv2.tv_usec = (tv1.tv_usec + u*1000) % 1000000;

						fprintf(stderr, "%s\tStart: %ld:%06ld -> %02d:%02d:%02d.%03d\n", scr_name, tv1.tv_sec, tv1.tv_usec, h, m, s, u);
						fprintf(stderr, "%s\tEnd  : %ld:%06ld\n\n", scr_name, tv2.tv_sec, tv2.tv_usec);

						if ( scr_cback[i].cback ) scr_cback[i].cback(c);
						
						while ( 1 ) {
							usleep(1);
							gettimeofday(&tv_tmp, NULL);
							if ( tv_tmp.tv_sec < tv2.tv_sec ) {
								continue;
							} else {
								if ( tv_tmp.tv_sec == tv2.tv_sec ) {
									if ( tv_tmp.tv_usec < tv2.tv_usec )
										continue;
								}
								break;
							}
						}
					}
#endif /* ABSOLUTE_TIME */
					break;
				}
			}
			
			if ( !scr_cback[i].token )
				scr_warn("Unknown token");

			if ( !skip_line() ) break;
		}
	}

	canon_exit_control(c);
}

int main( int argc, char **argv) {
	camera *cameras;

//	iso = argv[1];
	printf("Multican v0.0.4, by Jindrich Novy (jnovy@users.sourceforge.net)\n\n");

	if ( !canon_scan(&cameras) ) {
		fprintf(stderr, "No known Canon camera recognized.\n");
		return 1;
	}

	if ( !canon_init(cameras) ) {
		fprintf(stderr, "No recognized camera was successfully initialized.\n");
		return 1;
	}

	if ( argc > 1 ) {
		int i;
		char ch;
		char *scrs[USB_MAX_DEVICES];
		char *scr_names[USB_MAX_DEVICES];
		off_t scr_sizes[USB_MAX_DEVICES];
		camera *cams[USB_MAX_DEVICES];

		printf("User request to control %d camera%s.\n\n", argc-1, argc-1>1?"s":"");
		
		if ( argc-1 > USB_MAX_DEVICES ) {
			fprintf(stderr, "More scripts than the maximum possible USB devices specified.\n");
			exit(1);
		}

		for ( i=1; i<argc ; i++ ) {
			FILE *f = fopen(argv[i],"rt");
			int n, l;
			unsigned serno;
			
			scr_name = argv[i];
			scr_names[i-1] = scr_name;

			if ( !f )
				scr_err("Unable to open script");

			l = strlen(scr_name);

			if ( l < 10 )
				scr_err("Invalid script name -> must contain serial number of a camera at the end");

			for ( n=l-10; n<l; n++ )
				if ( scr_name[n] < '0' && scr_name[n] > '9' )
					scr_err("Serial number in script name contains invalid character");

			if ( sscanf(&scr_name[l-10], "%u", &serno) != 1 )
				scr_err("Bad serial number in script name");

			fseek(f, 0, SEEK_END);
			scr_sizes[i-1] = ftell(f);
			fseek(f, 0, SEEK_SET);

			scrs[i-1] = malloc(scr_sizes[i-1]);

			if (!fread(scrs[i-1], 1, scr_sizes[i-1], f))
				scr_err("Error reading script");

			fclose(f);

			for ( n=0; n<USB_MAX_DEVICES; n++ ) {
				camera *c = &cameras[n];

				if ( !c->ready ) continue;
		
				canon_identify_camera(c);
				canon_eos_id(c);

				if ( c->body_id == serno ) {
					putchar('\n');
					cams[i-1] = c;
					break;
				}
			}

			if ( n >= USB_MAX_DEVICES )
				scr_err("No active device matches serial number in script");
		}

		printf("\nPress ENTER to start the scripted sequence.\n");
		fscanf(stdin, "%c", &ch);
		printf("Sequence started...\n");

		{
			pid_t pids[USB_MAX_DEVICES];
			int totpids = 0, status = 0;
#ifdef ABSOLUTE_TIME
			gettimeofday(&start_tv, NULL);
#endif
			for ( i=1; i<argc; i++ ) {
				if ( !(pids[totpids++]=fork()) ) {
					scr_name = scr_names[i-1];
					scr_size = scr_sizes[i-1];
					scr = scrs[i-1];

					process_script( cams[i-1] );

					return 0;
				}
			}

			for ( i=0; i<totpids; i++ ) {
				struct sched_param schp;
				
				schp.sched_priority = SCHED_RR;
			
				sched_setscheduler(pids[i], SCHED_RR, &schp);
				waitpid(pids[i], &status, WUNTRACED);

				if ( !WIFEXITED(status) ) {
					scr_name = scr_names[i];
					scr_warn("Script failed");
				}
			}
		}
	} else
	{
		int i;

		tolerate_error = 1;

		for ( i=0; i<USB_MAX_DEVICES; i++ ) {
			camera *c = &cameras[i];

			if ( !c->ready ) continue;

			canon_identify_camera(c);
			canon_power_status(c);
			if ( !IS_TYPE(c,"1D") ) canon_eos_id(c);
//			canon_get_zoom(c);

			canon_init_control(c);
			canon_get_rel_params(c);
			canon_tv(c, "0\"5");
			canon_av(c, "5.6");
			canon_iso(c, "200");
			canon_set_rel_params(c);

//			canon_flash(c, FLASH_OFF);
			canon_transfer(c, FULL_TO_DRIVE);
//			canon_transfer(c, THUMB_TO_PC);
//			canon_avail_shots(c);

			canon_release_shutter(c);

			sleep(3);

			canon_get_rel_params(c);
			canon_tv(c, "0\"4");
			canon_av(c, "6.3");
			canon_iso(c, "100");
			canon_set_rel_params(c);
			
			canon_release_shutter(c);

			sleep(3);

			canon_get_rel_params(c);
			canon_tv(c, "0\"3");
			canon_av(c, "7.1");
			canon_iso(c, "400");
			canon_set_rel_params(c);

			canon_release_shutter(c);

			sleep(3);

			canon_exit_control(c);
		}
	}

	canon_close(&cameras);

	return 0;
}
